/*
 * COPYRIGHT (C) 2008, CANTATA
 *
 * CANTATA-RACE/RMC:
 * rmc.h
 *
 * DESCRIPTION:
 * Resource Manager Component (RMC) interface
 * 
 * The Resource Manger Component is a part of the CANTATA Runtime Environment. 
 * It is responsible for managing the memory for the components 1. It provides 
 * interfaces to request and discard memory budgets, and to allocate and free 
 * memory within an allotted budget.
 *
 * The RMC provides guaranteed memory access by granting memory budgets to 
 * components. Upon a budget request the RMC checks whether there is sufficient 
 * space to accommodate it. The component then can allocate memory within the 
 * budget. Upon each allocation request the RMC checks whether there is enough 
 * space available in the budget and makes sure that no other component “steels”
 * the allocated budget. A component can be sure that what it has requested and
 * was granted by the RMC will indeed be available.
 *
 * Memory within the budgets can be freed, as can the complete budgets be
 * discarded. When memory within a budget is freed it can be reallocated by the
 * component (the owner of the budget). When a budget is discarded it can be
 * requested again, also by other components.
 * 
 * Budgets can be shared between components by wrapping them inside special
 * components, such as a buffer or Quality of Service component. These
 * components can request several budgets and then control the access to the
 * shared memory from other components, e.g. by synchronizing reading and
 * writing to a buffer, or reassigning budgets between components to control
 * the QoS.
 *
 * The intended use for the RMC component is the following:
 *
 * Initialization phase:
 * - The RMC is initialized with a call to RMC_init(), which allocates and
 *   initializes the internal RMC data structures. This is usually performed
 *   by the Runtime Environment. After the RMC is initialized each component
 *   can request (several) budgets with RMC_RequestMemoryBudget(). RMC_init()
 *   must be called before any other RMC interface method.
 *
 * Operation phase:
 * - Each component allocates memory within a budget with RMC_AllocateMemory().
 * - When it wants to change the allocation within a budget, e.g. due to a mode
 *   change, then it resets the complete budget with RMC_ResetMemoryBudget(),
 *   or frees a particular memory allocation within a budget with
 *   RMC_FreeMoemory(). It can then allocate the freed memory again.
 * - A component can also discard a complete budget with RMC_DiscardMemoryBudget()
 *   and in this way make the freed space available to other components, through
 *   calls to RMC_RequestMemoryBudget().
 * - A component can claim additional budgets during the operation phase
 *   similarly to the initialization phase, by calling RMC_RequestMemoryBudget().
 *
 * Finalization phase:
 * - The RMC_finalize() function frees all the memory, which was allocated by
 *   the RMC during the RMC_init() call. It is usually called by the Runtime
 *   Environment.
 *
 * HISTORY:
 * 2008-04-22   0.1  Mike Holenderski (TU/e)
 * 2008-05-05   0.2  Mike Holenderski (TU/e)
 * 2008-05-29   0.3  Mike Holenderski (TU/e)
 *                   - add dynamic memory allocation within budgets
 * 2008-06-09   0.4  Mike Holenderski (TU/e)
 *                   - fixed minor bugs
 *                   - added checking for invalid budget handles
 *                   - made the methods thread safe
 * 2008-07-08   0.6  Mike Holenderski (TU/e)
 *                   - fixed minor bugs
 *                   - fixed utilization computation
 * 2008-07-08   0.7  Mike Holenderski (TU/e)
 *                   - added budget validity checking in RMC_AllocateMemory()
 * 2008-07-08   0.8  Mike Holenderski (TU/e)
 *                   - fixed utilization computation
 * 2008-07-08   0.9  Mike Holenderski (TU/e)
 *                   - fixed annoying global variable name conflicts
 * 2009-07-25   1.0  Mike Holenderski (TU/e)
 *                   - added RMC_AllocationHandleType
 *                   - fixed debug warnings due to unsigned long vs. int
 *                   - changed API, specifying the max number of budgets/allocations,
 *                     rather than including these in the size parameter of the request
 *                   - added RMC_RoundBytesToWords()
 * 2009-08-24   1.1  Mike Holenderski
 *                   - fixed bug with initialization of budget with 0 size.
 */

#ifndef __CR_RMC__
#define __CR_RMC__

#ifdef __cplusplus
extern "C" {
#endif

#include "cr_types.h"
#include "cr_errors.h"


/*
 * Error codes, specific to RMC.
 */

#define CR_RMC_ERROR_FAIL               CR_ERROR_ASSIGN(0)
#define CR_RMC_ERROR_RESOURCE           CR_ERROR_ASSIGN(1)
#define CR_RMC_ERROR_BUDGET             CR_ERROR_ASSIGN(2)
#define CR_RMC_ERROR_RMC                CR_ERROR_ASSIGN(3)
#define CR_RMC_ERROR_MEMORY             CR_ERROR_ASSIGN(4)
#define CR_RMC_ERROR_INVALID_HEAD_SIZE  CR_ERROR_ASSIGN(5)

/*
 * The size of memory allocaltion units. Must be a multiple of the word size
 * typically 4. 
 */

#define RMC_UNIT_SIZE sizeof(int)

/*
 * Parameter names to the RMC_init() method.
 */
 
#define RMC_PARAM_MEMORY_SIZE "memory_size"
#define RMC_PARAM_MAX_NUMBER_OF_BUDGETS "budget_number"

/* 
 * Memory size, in bytes.
 */

typedef CR_UInt32 RMC_MemorySizeType;

/*
 * Handle to a budget. Handles are created and managed by the RMC. Implemented as
 * a pointer to some internal datastructure within RMC. To be used only as 
 * a unique budget identifier.
 */

typedef void* RMC_BudgetHandleType;

/*
 * Handle to an allocation. Handles are created and managed by the RMC. Implemented as
 * a pointer to some internal datastructure within RMC. To be used only as 
 * a unique budget identifier.
 */

typedef void* RMC_AllocationHandleType;

/*
 * Memory budget, specifies part of the memory budget, in particular
 * how much of which memory. Allocated budgets are contiguous chunks of memory.
 */

typedef struct RMC_MemoryBudgetRequestType
{
  RMC_MemorySizeType  allocationSize;  /* size in bytes. Note that it must include the
                                overhead of RMC_MEMORY_HEAD_SIZE per allocation
                                within the budget. */
  CR_UInt32 maxNumberOfAllocations; /* maximum number of allocations which will be requested
                                within the budget */
} RMC_MemoryBudgetRequestType;

/*
 * Memory request record (to allocate memory within a budget).
 * Specifies part of the memory budget, in particular how much of which memory.
 */

typedef struct RMC_MemoryRequestType
{
  RMC_MemorySizeType   size;   /* size in bytes */
  RMC_BudgetHandleType budget; /* handle to the budget where memory is to be
                                  allocated. */ 
} RMC_MemoryRequestType;

/* FUNCTION:
 * RMC_init
 *
 * DESCRIPTION:
 * Initializes the RMC component and its internal data structures. This method
 * must be called before calling any of the other RMC interface methods.
 * Note that this method is called by teh RTE, so a component must not call it!
 *
 * PARAMETERS:
 * (in)   pParameters     - creation parameters. Required parameters:
 *                          RMC_PARAM_MEMORY_SIZE 
 *
 * RETURN:
 * CR_RESULT_OK           - successfully initialized.
 * CR_ERROR_INVALID_PARAM - pParameters is NULL, or pParameters does not contain
 *                          RMC_PARAM_MEMORY_SIZE, or contains an invalid key
 * CR_RMC_ERROR_MEMORY    - There is not enough memory space to allocate RMC
 *                          according to the settings specified in rmc_config.h
 * CR_RMC_ERROR_RMC       - An internal data structure could not be initialized.
 * CR_RMC_ERROR_INVALID_HEAD_SIZE      
 *                        - The RMC_BUDGET_HEAD_SIZE or RMC_MEMORY_HEAD_SIZE do
 *                          not correspond to the sizes of the data strucutres
 *                          defined in rmc.c.
 */

extern CR_Result RMC_init(CR_CreationParameterType* pParameters);

/* FUNCTION:
 * RMC_finalize
 *
 * DESCRIPTION:
 * Finalize the RMC component and free its internal data structures. After 
 * calling this method, all data maanged by the RMC, e.g. memory budgets, is 
 * discarded.
 * Note that this method is called by teh RTE, so a component must not call it!
 *
 * RETURN:
 * CR_RESULT_OK           - successfully finalized.
 */

extern CR_Result RMC_finalize(void);


/* FUNCTION:
 * RMC_RequestMemoryBudget
 *
 * DESCRIPTION:
 * Resuest a memory budget. Before this method is called, the budget is
 * considered empty.
 *
 * PARAMETERS:
 * (in)   handle          - pointer to the object instance, used as a unique 
 *                          identifier for objects
 * (in)   pRequest        - pointer to the budget request
 * (out)  pBudgetHandle   - pointer to where to return the budget id
 *
 * RETURN:
 * CR_RESULT_OK             - budget can be guaranteed. pBudgetHandle identifies
 *                            the newly created budget.
 * CR_ERROR_INVALID_PARAM   - handle is NULL, or
 *                            pRequest is NULL or cannot be cast, or
 *                            pBudgetHandle is NULL
 * CR_RMC_ERROR_RESOURCE    - Not enough space in the memory to allocate 
 *                            budget. 
 * CR_RMC_ERROR_BUDGET      - Not enough space in the memory to allocate 
 *                            internal datastructures for the budget. 
 */

extern CR_Result RMC_RequestMemoryBudget(
                CR_ObjectType handle,
                RMC_MemoryBudgetRequestType* pRequest,
                RMC_BudgetHandleType* pBudgetHandle
);

/* FUNCTION:
 * RMC_DiscardMemoryBudget
 *
 * DESCRIPTION:
 * Discards the memory budget and makes it available to for other requests.
 * The available budget space may be combined with other available space, so that
 * a larger budget can be request, however, this is not guaranteed.
 *
 * PARAMETERS:
 * (in)   budgetHandle      - budget identifer, for the budget to be discarded.
 * 
 * RETURN:
 * CR_RESULT_OK             - budget was discarded.
 * CR_ERROR_INVALID_PARAM   - budgetHandle is NULL, or
 *                            budgetHandle is not a valid budget ID
 */

extern CR_Result RMC_DiscardMemoryBudget(
                RMC_BudgetHandleType budgetHandle
);

/* FUNCTION:
 * RMC_ResetMemoryBudget
 *
 * DESCRIPTION:
 * Resets the budget to the state after RMC_RequestMemoryBudget, i.e. memory can
 * be allocated from the beginning of the budget. All previously allocated 
 * memory requests are invalid.
 *
 * PARAMETERS:
 * (in)   budgetHandle      - budget identifier, for the budget to be reset.
 * 
 * RETURN:
 * CR_RESULT_OK             - budget was reset.
 * CR_ERROR_INVALID_PARAM   - BudgetHandle is NULL, or
 *                            budgetHandle is not a valid budget ID
 */

extern CR_Result RMC_ResetMemoryBudget(
                RMC_BudgetHandleType budgetHandle
);

/* FUNCTION:
 * RMC_AllocateMemory
 *
 * DESCRIPTION:
 * Allocate part of the memory budget. The allocated memory can be treated as 
 * one contiguous chunk.
 *
 * PARAMETERS:
 * (in)   pRequest          - pointer to the memory request
 * (out)  allocationHandle  - start address of the allocated memory chunk
 * 
 * RETURN: 
 * CR_RESULT_OK             - requested memory was allocated. base points the 
 *                            beginning of the allocation. 
 * CR_ERROR_INVALID_PARAM   - pRequest is NULL, or ppBase is NULL, 
 *                            or pRequest.budget is NULL, or pRequest.size < 1
 * CR_RMC_ERROR_BUDGET      - request could not be fulfilled, e.g. trying to 
 *                            allocate more than agreed in the budget.
 */

extern CR_Result RMC_AllocateMemory(
                RMC_MemoryRequestType* pRequest,
                RMC_AllocationHandleType* pAllocationHandle
);

/* FUNCTION:
 * RMC_FreeMemory
 *
 * DESCRIPTION:
 * Free part of the memory budget. Frees all memory which was previously 
 * allocated starting at the address refered to by base.
 *
 * PARAMETERS:
 * (in)   budgetHandle      - pointer to the object instance
 * (in)   allocationHandle  - start address of the memory chunk to be freed
 * 
 * RETURN: 
 * CR_RESULT_OK             - requested memory was freed 
 * CR_ERROR_INVALID_PARAM   - budgetHandle is NULL, or base is NULL
 * CR_RMC_ERROR_BUDGET      - trying to free a wrong memory location, e.g. not 
 *                            assigned to this budget
 */

extern CR_Result RMC_FreeMemory(
                RMC_BudgetHandleType budgetHandle,
                RMC_AllocationHandleType allocationHandle
);

/* FUNCTION:
 * RMC_getAvailableBudget
 *
 * DESCRIPTION:
 * Get the number of bytes currently free within the budget. Note that since
 * the RMC does not address fragmentation, the number of bytes returned by this
 * function may not be allocable in one contiguous chunk.
 *
 * PARAMETERS:
 * (in)   budgetHandle      - pointer to the object instance
 * (out)  available        - pointer to store the number of available bytes
 * 
 * RETURN: 
 * CR_RESULT_OK             - available was successfully returned 
 * CR_ERROR_INVALID_PARAM   - budgetHandle is NULL, or available is NULL, or
 *                            budgetHandle is not a valid budget ID
 */

extern CR_Result RMC_getAvailableBudget(
                          RMC_BudgetHandleType budgetHandle,
                          CR_UInt32* available);

/* FUNCTION:
 * RMC_getAvailableMemory
 *
 * DESCRIPTION:
 * Get the total number of bytes currently available to new budgets requests. 
 * Note that since the RMC does not address fragmentation, the number of bytes 
 * returned by this fubction may not be allocatable in one contiguous chunk.
 *
 * PARAMETERS:
 * (out)  available        - pointer to store the number of available bytes
 * 
 * RETURN: 
 * CR_RESULT_OK             - available was successfully returned 
 * CR_ERROR_INVALID_PARAM   - available is NULL
 */

extern CR_Result RMC_getAvailableMemory(CR_UInt32* available);

/* FUNCTION:
 * RMC_RoundBytesToWords
 *
 * DESCRIPTION:
 * A utility function rounding up number of bytes to an integral number of words.
 *
 * PARAMETERS:
 * (in)  bytes        - number of bytes
 * 
 * RETURN: 
 * number of bytes equal or larger than the input number of bytes, and a multiple of
 * the word size.
 */

extern CR_UInt32 RMC_RoundBytesToWords(CR_UInt32 bytes);

#ifdef __cplusplus
}
#endif

#endif /* __CR_RMC__ */