/*
 * COPYRIGHT (C) 2008, CANTATA
 *
 * CANTATA-RACE/RMC:
 * rmc.c
 *
 * DESCRIPTION:
 * Resource Manager Component (RMC) implementation
 *
 * See rmc.h for a description of the RMC.
 *
 * HISTORY:
 * 2008-04-22   0.1  Mike Holenderski (TU/e)
 * 2008-05-05   0.2  Mike Holenderski (TU/e)
 * 2008-05-29   0.3  Mike Holenderski (TU/e)
 *                   - add dynamic memory allocation within budgets
 * 2008-06-09   0.4  Mike Holenderski (TU/e)
 *                   - fixed minor bugs
 *                   - added checking for invalid budget handles
 *                   - made the methods thread safe
 * 2008-07-08   0.6  Mike Holenderski (TU/e)
 *                   - fixed minor bugs
 *                   - fixed utilization computation
 * 2008-07-08   0.7  Mike Holenderski (TU/e)
 *                   - added budget validity checking in RMC_AllocateMemory()
 * 2008-07-08   0.8  Mike Holenderski (TU/e)
 *                   - fixed utilization computation
 * 2008-07-08   0.9  Mike Holenderski (TU/e)
 *                   - fixed annoying global variable name conflicts
 * 2009-07-25   1.0  Mike Holenderski (TU/e)
 *                   - added RMC_AllocationHandleType
 *                   - fixed debug warnings due to unsigned long vs. int
 *                   - changed API, specifying the max number of budgets/allocations,
 *                     rather than including these in the size parameter of the request
 *                   - added RMC_RoundBytesToWords()
 * 2009-08-24   1.1  Mike Holenderski
 *                   - fixed bug with initialization of budget with 0 size.
 */

/* make sure that the configurations are defined */
#define __CR_RMC_DEFINE_CONFIG__

#include "rmc.h"
#include "rmc_debug.h"
#include "rmc_types.h"
#include <stdlib.h>
#include <string.h>

/* the OSAL framwork */
#ifndef CR_NO_OSAL
#include "common_types.h"
#include "osapi.h"
#endif

static const CR_UInt32 RMC_BUDGET_HEAD_SIZE = sizeof(RMC_MemoryBudgetHeadType);
static const CR_UInt32 RMC_MEMORY_HEAD_SIZE = sizeof(RMC_MemoryAllocationHeadType);

/*
 * General type for the nodes in a free list. It is a sort of super class for
 * memory allocation info and memory budget info. It allows to reuse functions
 * for allocating and discarding budgets and memory.
 *
 * IMPORTANT: make sure that the first two fields in the budget and memory info
 * records are the same as for RMC_NodeType.
 */
 
typedef struct _RMC_NodeType RMC_NodeType;

struct _RMC_NodeType {
	RMC_MemorySizeType size;
  RMC_NodeType *next;
};

/*
 * Global data structures keeping track of all available and allocated budgets.
 */

static RMC_MemoryBudgetHeadType* RMC_TheBudgetFreeList;
RMC_MemoryBudgetHeadType* RMC_budgetFreeList;
RMC_MemoryBudgetHeadType* RMC_budgetAllocatedList;

/*
 * Memory size available to the RMC.
 */

static CR_UInt32 RMC_MemorySize = 0;

/*
 * Total size of allocated memory budgets.
 */

static CR_UInt32 RMC_MemoryBudgetUtilization = 0;

/*
 * Mutex for synchronizing access to the RMC interface methods.
 */

#ifndef CR_NO_OSAL
#define RMC_mutex_id mutex_id
static uint32 RMC_mutex_id;
#endif

/*
 * Initialize the free list in a budget. The size and next field must be
 * specified.
 */

void initBudget(RMC_MemoryBudgetHeadType* pBudget) {
  if (pBudget->size > 0) {
    pBudget->freeList = (RMC_MemoryAllocationHeadType*)(pBudget + 1);
    pBudget->freeList->size = pBudget->size - RMC_MEMORY_HEAD_SIZE;
    //pBudget->freeList->size = pBudget->size;
    pBudget->freeList->next = NULL;
    pBudget->allocatedList = NULL;
    pBudget->utilization = 0;
  } else {
    pBudget->freeList = NULL;
    pBudget->allocatedList = NULL;
    pBudget->utilization = 0;
  }
}

/*
 * Check if target node is a member of a list.
 *
 * Returns:
 * 1 if the list starting at pBegin contains pTarget
 * 0 if the list starting at pBegin does not contain pTarget, or pBegin == NULL
 *   or pTarget == NULL
 */

int inList(RMC_NodeType* pTarget, RMC_NodeType* pBegin) {
  for (;pBegin != NULL; pBegin = pBegin->next) {
    if (pBegin == pTarget) {
      return 1;
    }
  }
  
  return 0;
}

/*
 * Remove a node from allocated list.
 */

int removeFromList(RMC_NodeType* pTarget, RMC_NodeType** pBegin) {
  RMC_NodeType *p, *oldp;
  for (p = *pBegin, oldp = NULL; p != NULL; oldp = p, p = p->next) {
    if (oldp != NULL && p == pTarget) {
      /* remove pTarget from the list */
      oldp->next = p->next;
      return 1;
    } else if (oldp == NULL && p == pTarget) {
      *pBegin = p->next;
      return 1;
    }
  }
  
  return 0;
}

/*
 * Allocate a budget within a list of available budgets, or memory within a list
 * of a vailable memory allocations of a budget.
 */
 
RMC_NodeType* allocateMemory( CR_UInt32 size, 
                              CR_UInt32 headSize, 
                              CR_UInt32 unitSize, 
                              RMC_NodeType** pBegin,
                              RMC_NodeType** pBeginAllocated,
                              CR_UInt32* pUtilization
                              ) {
  
	RMC_NodeType *p, *q, *oldp;
	CR_UInt32 tailSize, totalSize;
  
  /* round up node tail to an integral part of unit size */
	tailSize = ((size + unitSize - 1) / unitSize) * unitSize;
  totalSize = headSize + tailSize;

  debug("allocate(): tailSize  =%lu\n", tailSize);
  debug("allocate(): headSize  =%lu\n", headSize);
  debug("allocate(): totalSize =%lu\n", totalSize);
  
	/* See if we can satisfy the request from the freelist. */
	for (p = *pBegin, oldp = NULL;
	     p != NULL;
	     oldp = p, p = p->next) {
		if (p->size >= totalSize) {
      debug("allocate(): fits within a freelist (%lu <= %lu)\n", totalSize, p->size);

      /* node is larger, shorten, and return the tail */
      p->size -= totalSize; 
			q = (RMC_NodeType*)((CR_UInt32)p + headSize + p->size);
      q->size = tailSize;
      
      /* add to the beginning of the allocated list */
      q->next = *pBeginAllocated;
      *pBeginAllocated = q;
      
      /* update utilization by adding size of head and tail */
      *pUtilization = *pUtilization + totalSize;
      
			return q;
		}
    /* if it does not fit into the freelist, check if it fits in the total node 
       size, including head */
		if (p->size >= tailSize) {
      debug("allocate(): tail fits (almost) exactly in a freelist (%lu <= %lu)\n", tailSize, p->size);

      /* (almost) exact match, remove from freelist */
			if (oldp == NULL) {
				*pBegin = p->next;
			} else {
				oldp->next = p->next;
      }
      
      /* to avoid creating gaps, take the old size, even if it is larger than
         requested */
      p->next = *pBeginAllocated;
      *pBeginAllocated = p;
         
      /* update utilization by adding size of the occupied node */
      *pUtilization = *pUtilization + headSize + p->size;   
      
      return p;
		}
	}

  debug("allocate(): does not fit\n");
  return NULL;
}

/*
 * Free previousely allocated memory.
 *
 * Traverse the free list of the budget until a free chunk after the allocation
 * to be freed. Insert a new free chunk in the free list at this position and 
 * merge with neighboring chunks, in case they are free. This way teh free list
 * is sorted and allows merging of free chunks.
 */

int freeMemory(RMC_NodeType* pTarget, RMC_NodeType** pBegin, RMC_NodeType** pBeginAllocated) {
	RMC_NodeType *p,*oldp;
  
  /* remove the target node form teh allocated list */
  removeFromList(pTarget, pBeginAllocated);
  
  /* insert the target node into the free list. traverse the freelist looking 
     for where the pTarget would fit to keep it sorted */
  for (p = *pBegin, oldp = 0;
	     p != NULL;
	     oldp = p, p = p->next) {
		if (pTarget < p) {
      debug("freeMemory(): found a spot\n");

      /* check if inserting at the beginning of the free list */
      if (oldp == NULL) {
        debug("freeMemory(): insert allocation at the beginning of the free list\n");
        *pBegin = pTarget;
        pTarget->next = p;
      } else {
        debug("freeMemory(): insert allocation between %p and %p\n", (void*)oldp, (void*)p);
        oldp->next = pTarget;
        pTarget->next = p;
      }
      
      return 1;
		}
		if (pTarget == p) {
      debug("freeMemory(): the allocation is already free\n");
			return 1;
		}
	}
        
  debug("freeMemory(): insert allocation at the end of the free list\n");
  if (oldp == NULL) {
    *pBegin = pTarget;
  } else {
    oldp->next = pTarget;
  }
  pTarget->next = NULL;
  
  return 1;
}

/*
 * Merge neighbouring nodes in a free list. The nodes can be either budgets or
 * memory allcoations within the free list of a budget.
 *
 * All nodes in the list are traversed and joined, in case one ends where the 
 * next one begins.
 *
 * Returns number of merged nodes.
 */
 
CR_UInt32 mergeMemory(RMC_NodeType* pBegin, CR_UInt32 headSize) {
	RMC_NodeType *p, *oldp;
  CR_UInt32 merged;
  
  p = pBegin;
  oldp = 0;
	merged = 0;
  while (p) {
		if ((CR_UInt32)p + headSize + p->size == (CR_UInt32)p->next) {
      debug("merge(): merging (%p, %lu) with (%p, %lu)\n", (void*)p, p->size, (void*)p->next, p->next->size);
      p->size = p->size + headSize + p->next->size;
      p->next = p->next->next;
      merged++;
    } else {
      oldp = p;
      p = p->next;
    }
  }
  return merged;
}

/*
 * Initialize the RMC and its global list of available budgets. Note that it is 
 * assumed that RMC_init() is called only once, before RMC_finalize().
 */

CR_Result RMC_init(CR_CreationParameterType* pParameters) {
  CR_UInt16 i;
  CR_UInt32 size;
  CR_UInt32 maxNumberOfBudgets;

  /* check if the head sizes defined in rmc.h correspond to the real head sizes */
  if (RMC_BUDGET_HEAD_SIZE != sizeof(RMC_MemoryBudgetHeadType)) {
    return CR_RMC_ERROR_INVALID_HEAD_SIZE;
  }
  if (RMC_MEMORY_HEAD_SIZE != sizeof(RMC_MemoryAllocationHeadType)) {
    return CR_RMC_ERROR_INVALID_HEAD_SIZE;
  }

  /* parse the parameters */
  if (pParameters == NULL) {
    return CR_ERROR_INVALID_PARAM;
  }

  for (i = 0; CR_NULL != pParameters[i].Key; i++) {
    if (strcmp( pParameters[i].Key, RMC_PARAM_MEMORY_SIZE ) == 0) {
      size = (CR_UInt32) atol( pParameters[i].Value );
    } else if (strcmp( pParameters[i].Key, RMC_PARAM_MAX_NUMBER_OF_BUDGETS ) == 0) {
      maxNumberOfBudgets = (CR_UInt32) atol( pParameters[i].Value );
    } else {
      return CR_ERROR_INVALID_PARAM;
    }
  }
  
  RMC_MemorySize = size + maxNumberOfBudgets*RMC_BUDGET_HEAD_SIZE;

  /* check if there is anything to allocate */
  if (RMC_MemorySize == 0) {
    return CR_ERROR_INVALID_PARAM;
  }

  /* allocate the memory which will be managed by the RMC */
  if ((RMC_budgetFreeList = (RMC_MemoryBudgetHeadType*) malloc(RMC_MemorySize) ) == NULL) {
    return CR_RMC_ERROR_MEMORY;
	}
  
  /* save the budetFreeList in an untouched pointer, to free it during finalization */
  RMC_TheBudgetFreeList = RMC_budgetFreeList;
  
  /* initialize some internal data structures: free list and allocated list */
  RMC_budgetFreeList->size = RMC_MemorySize - RMC_BUDGET_HEAD_SIZE;
  RMC_budgetFreeList->next = NULL;
  initBudget(RMC_budgetFreeList);

  RMC_budgetAllocatedList = NULL;
  
  /* craete the mutex for synchronizeing access to the RMC interface methods */
  #ifndef CR_NO_OSAL
    if (OS_MutSemCreate( &RMC_mutex_id, "RMC_Mutex", 0) != OS_SUCCESS) {
      return CR_RMC_ERROR_RMC;
    }
  #endif
  
  return CR_RESULT_OK;
}

/*
 * Finalize the RMC and free any memory used by the budgets.
 */
 
CR_Result RMC_finalize(void) {
  /* avoid double freeing of the RMC_TheBudgetFreeList */
  if (RMC_TheBudgetFreeList == NULL) {
    return CR_RMC_ERROR_RMC;
  }

  /* free TheBudgetFreeList, instead of budgetFreeList, to prevent memory leak */
  RMC_budgetAllocatedList = NULL;
  free(RMC_TheBudgetFreeList);
  RMC_TheBudgetFreeList = NULL;
  
  /* delete the mutex */
  #ifndef CR_NO_OSAL
    if (OS_MutSemDelete(RMC_mutex_id) != OS_SUCCESS) {
      return CR_RMC_ERROR_RMC;
    }
  #endif
  
  return CR_RESULT_OK;
}

/*
 * RMC_RequestMemoryBudget
 */
 
CR_Result RMC_RequestMemoryBudget(
                CR_ObjectType handle,
                RMC_MemoryBudgetRequestType* pRequest,
                RMC_BudgetHandleType* pBudgetHandle
) {
  RMC_MemoryBudgetHeadType* pBudget;
  CR_Int32 size;

  /* check if parameters are not NULL */
  if (handle == NULL || pRequest == NULL || pBudgetHandle == NULL) {
    return CR_ERROR_INVALID_PARAM;
  }

  debug("RMC_RequestMemoryBudget(): allocationSize=%lu allocationNumber=%lu\n", pRequest->allocationSize, pRequest->maxNumberOfAllocations);

  #ifndef CR_NO_OSAL
    OS_MutSemTake(RMC_mutex_id);
  #endif
  
  /* add the header overhead */
  size = pRequest->allocationSize + pRequest->maxNumberOfAllocations*RMC_MEMORY_HEAD_SIZE;
    
  /* allocate the budget within the freelist of budgets. Use allocateMemory()
     customized for budgets */
  pBudget = (RMC_MemoryBudgetHeadType*) allocateMemory(
                size,
                RMC_BUDGET_HEAD_SIZE, RMC_UNIT_SIZE,
                (RMC_NodeType**)&RMC_budgetFreeList,
                (RMC_NodeType**)&RMC_budgetAllocatedList,
                &RMC_MemoryBudgetUtilization
                );

  /* Check if enough space was available to allocate a new budget */
	if (pBudget == NULL) {
    #ifndef CR_NO_OSAL
      OS_MutSemGive(RMC_mutex_id);  
    #endif
    
		return CR_RMC_ERROR_BUDGET;
	}

  pBudget->owner = handle;
  pBudget->maxNumberOfAllocations = pRequest->maxNumberOfAllocations;

  /* initialize the free list within the budget */
  initBudget(pBudget);

  #ifndef CR_NO_OSAL
    OS_MutSemGive(RMC_mutex_id);  
  #endif
      
  debugb(pBudget, "RMC_RequestMemoryBudget(): pBudget");
  
  /* budget handle is a pointer to the budget. This way we guarantee that we do
     not run out of handle ids (only one budget can start at a particular 
     address). */
  *pBudgetHandle = (RMC_BudgetHandleType)pBudget;
    
  return CR_RESULT_OK;
}

/*
 * RMC_DiscardMemoryBudget
 */
 
CR_Result RMC_DiscardMemoryBudget(
                RMC_BudgetHandleType budgetHandle
) {
  RMC_MemoryBudgetHeadType* pBudget;
    
  /* check if parameters are not NULL */
  if (budgetHandle == NULL) {
    return CR_ERROR_INVALID_PARAM;
  }

  pBudget = (RMC_MemoryBudgetHeadType*)budgetHandle;

  #ifndef CR_NO_OSAL
    OS_MutSemTake(RMC_mutex_id);
  #endif
  
  /* check if budgetHandle is a valid budget pointer */
  if (!inList((RMC_NodeType*)pBudget, (RMC_NodeType*)RMC_budgetAllocatedList)) {
    #ifndef CR_NO_OSAL
      OS_MutSemGive(RMC_mutex_id);  
    #endif

    return CR_ERROR_INVALID_PARAM;
  }

  freeMemory( (RMC_NodeType*)pBudget, 
              (RMC_NodeType**)&RMC_budgetFreeList, 
              (RMC_NodeType**)&RMC_budgetAllocatedList
              );

  /* subtract from total budget utilization. Do it before merging, because it
   * may change the pBudget->size */
  RMC_MemoryBudgetUtilization -= (RMC_BUDGET_HEAD_SIZE + pBudget->size);

  mergeMemory((RMC_NodeType*)RMC_budgetFreeList, RMC_BUDGET_HEAD_SIZE);
  
  
  #ifndef CR_NO_OSAL
    OS_MutSemGive(RMC_mutex_id);  
  #endif
    
  return CR_RESULT_OK;
}

/*
 * RMC_ResetMemoryBudget
 */

CR_Result RMC_ResetMemoryBudget(
                RMC_BudgetHandleType budgetHandle
) {
  RMC_MemoryBudgetHeadType* pBudget;
  
  /* check if parameters are not NULL */
  if (budgetHandle == NULL) {
    return CR_ERROR_INVALID_PARAM;
  }

  pBudget = (RMC_MemoryBudgetHeadType*)budgetHandle;

  #ifndef CR_NO_OSAL
    OS_MutSemTake(RMC_mutex_id);
  #endif
  
  /* check if budgetHandle is a valid budget pointer */
  if (!inList((RMC_NodeType*)pBudget, (RMC_NodeType*)RMC_budgetAllocatedList)) {
    #ifndef CR_NO_OSAL
      OS_MutSemGive(RMC_mutex_id);  
    #endif
    
    return CR_ERROR_INVALID_PARAM;
  }

  /* reset the free list and utilization */
  initBudget(pBudget);

  #ifndef CR_NO_OSAL
    OS_MutSemGive(RMC_mutex_id);  
  #endif
  
  /* check if budget */
  return CR_RESULT_OK;
}

/*
 * RMC_AllocateMemory
 */

CR_Result RMC_AllocateMemory(
                RMC_MemoryRequestType* pRequest,
                RMC_AllocationHandleType* pAllocationHandle
) {
  RMC_MemorySizeType size;
  
  /* pointer to the budget informatioin */
  RMC_MemoryBudgetHeadType* pBudget;
  
  /* pointer to the memory allocation */
  RMC_MemoryAllocationHeadType* pAllocation;
  
  /* check if parameters are not NULL */
  if (pRequest == NULL || pAllocationHandle == NULL || pRequest->budget == NULL) {
    return CR_ERROR_INVALID_PARAM;
  }

  debug("RMC_AllocateMemory(): pRequest.size=%lu\n", pRequest->size);
  
  /* check if the size is legal */
  if (pRequest->size < 1) {
    return CR_ERROR_INVALID_PARAM;  
  }
  
  /* cast the budgetHandle to budget info record pointer */
  pBudget = (RMC_MemoryBudgetHeadType*)pRequest->budget;

  #ifndef CR_NO_OSAL
    OS_MutSemTake(RMC_mutex_id);
  #endif

  /* check if budgetHandle is a valid budget pointer */
  if (!inList((RMC_NodeType*)pBudget, (RMC_NodeType*)RMC_budgetAllocatedList)) {
    #ifndef CR_NO_OSAL
      OS_MutSemGive(RMC_mutex_id);  
    #endif

    return CR_ERROR_INVALID_PARAM;
  }
  
  size = pRequest->size;
  
  /* allocate the memory within the budget's freelist. Use allocateMemory()
     customized for memory allocation */
  pAllocation = (RMC_MemoryAllocationHeadType*)allocateMemory(
                      size, 
                      RMC_MEMORY_HEAD_SIZE, RMC_UNIT_SIZE, 
                      (RMC_NodeType**)&(pBudget->freeList),
                      (RMC_NodeType**)&(pBudget->allocatedList),
                      &pBudget->utilization
                      );

  /* check if could allocate, i.e. if the returned pointer is not NULL */
  if (pAllocation) {

    #ifndef CR_NO_OSAL
      OS_MutSemGive(RMC_mutex_id);  
    #endif
    
    *pAllocationHandle = (RMC_AllocationHandleType)(pAllocation + 1);

    return CR_RESULT_OK;
  } else {  
    #ifndef CR_NO_OSAL
      OS_MutSemGive(RMC_mutex_id);  
    #endif

    *pAllocationHandle = NULL;
    return CR_RMC_ERROR_BUDGET;
  }
}

/*
 * RMC_FreeMemory
 *
 * Currently no memory management is supported for dealing with fragmentation. 
 * Freed memory is simply added to the free list of the corresponding budget.
 * Neighboring nodes in the free list which occupy consecutive memory space
 * are merged to make larger available space allocations.
 */

CR_Result RMC_FreeMemory(
                RMC_BudgetHandleType budgetHandle,
                RMC_AllocationHandleType allocationHandle
) {
  RMC_MemoryBudgetHeadType* pBudget;
  RMC_MemoryAllocationHeadType* pAllocation;
  
  /* check if parameters are not NULL */
  if (budgetHandle == NULL || allocationHandle == NULL) {
    return CR_ERROR_INVALID_PARAM;
  }

  pBudget = (RMC_MemoryBudgetHeadType*)budgetHandle;
  pAllocation = (RMC_MemoryAllocationHeadType*)allocationHandle - 1;
  
  #ifndef CR_NO_OSAL
    OS_MutSemTake(RMC_mutex_id);
  #endif

  /* check if budgetHandle is a valid budget pointer */
  if (!inList((RMC_NodeType*)pBudget, (RMC_NodeType*)RMC_budgetAllocatedList)) {
    #ifndef CR_NO_OSAL
      OS_MutSemGive(RMC_mutex_id);  
    #endif

    return CR_ERROR_INVALID_PARAM;
  }

  /* check if allocationHandle is a valid memory allocation within pBudget */
  if (!inList((RMC_NodeType*)pAllocation, (RMC_NodeType*)pBudget->allocatedList)) {
    #ifndef CR_NO_OSAL
      OS_MutSemGive(RMC_mutex_id);  
    #endif

    return CR_ERROR_INVALID_PARAM;
  }

  debugFreeList(pBudget, "RMC_FreeMemory(): freeList(before)");

  freeMemory((RMC_NodeType*)pAllocation, (RMC_NodeType**)&(pBudget->freeList), (RMC_NodeType**)&(pBudget->allocatedList));

  /* subtract from the budget utilization. Do it before merging, because it
   * may change the pAllocation->size */
  pBudget->utilization -= (RMC_MEMORY_HEAD_SIZE + pAllocation->size);

  mergeMemory((RMC_NodeType*)pBudget->freeList, RMC_MEMORY_HEAD_SIZE);

  debugFreeList(pBudget, "RMC_FreeMemory(): freeList(after)");
  
  #ifndef CR_NO_OSAL
    OS_MutSemGive(RMC_mutex_id);  
  #endif
  
  return CR_RESULT_OK;
}

/*
 * RMC_getAvailableMemory
 */

CR_Result RMC_getAvailableMemory(CR_UInt32* available) {  
  /* check if parameters are not NULL */
  if (available == NULL) {
    return CR_ERROR_INVALID_PARAM;
  }
  
  *available = RMC_MemorySize - RMC_MemoryBudgetUtilization;
  return CR_RESULT_OK;
}

/*
 * RMC_getAvailableBudget
 */

CR_Result RMC_getAvailableBudget(RMC_BudgetHandleType budgetHandle, CR_UInt32* available) {
  RMC_MemoryBudgetHeadType* pBudget;
  RMC_MemoryAllocationHeadType* p;
  CR_UInt32 i;
  
  /* check if parameters are not NULL */
  if (budgetHandle == NULL || available == NULL) {
    return CR_ERROR_INVALID_PARAM;
  }

  pBudget = (RMC_MemoryBudgetHeadType*)budgetHandle;

  #ifndef CR_NO_OSAL
    OS_MutSemTake(RMC_mutex_id);
  #endif

  /* check if budgetHandle is a valid budget pointer */
  if (!inList((RMC_NodeType*)pBudget, (RMC_NodeType*)RMC_budgetAllocatedList)) {
    #ifndef CR_NO_OSAL
      OS_MutSemGive(RMC_mutex_id);  
    #endif
  
    return CR_ERROR_INVALID_PARAM;
  }
  
  /* get number of allocations */
  i = 0;
  for (p = pBudget->allocatedList; p != NULL; p = p->next) {
    i++;
  }
  
  if (i > pBudget->maxNumberOfAllocations)
    debug("more allocaitons %lu > %lu", i , pBudget->maxNumberOfAllocations);
  
  *available = pBudget->size - pBudget->utilization - (pBudget->maxNumberOfAllocations - i)*RMC_MEMORY_HEAD_SIZE;
    
  #ifndef CR_NO_OSAL
    OS_MutSemGive(RMC_mutex_id);  
  #endif
    
  return CR_RESULT_OK;
}

/*
 * RMC_RoundBytesToWords
 */

CR_UInt32 RMC_RoundBytesToWords(CR_UInt32 bytes) {
  return (bytes/sizeof(void*))*sizeof(void*) + (bytes % sizeof(void*) ? sizeof(void*) : 0);
}
