/*
 * COPYRIGHT (C) 2008, CANTATA
 *
 * CANTATA-RACE/RMC:
 * test_rmc.c
 *
 * DESCRIPTION:
 * Resource Manager Component (RMC) test suite.
 *
 * See rmc.h for a description of the RMC.
 *
 * HISTORY:
 * 2008-04-25   0.1  Mike Holenderski (TU/e)
 * 2008-05-05   0.2  Mike Holenderski (TU/e)
 */

#include "rmc.h"
#include "rmc_types.h"

#include "test_rmc.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

static const CR_UInt32 RMC_BUDGET_HEAD_SIZE = sizeof(RMC_MemoryBudgetHeadType);
static const CR_UInt32 RMC_MEMORY_HEAD_SIZE = sizeof(RMC_MemoryAllocationHeadType);

/* references to the global variables in the RMC impolementation */
extern  RMC_MemoryBudgetHeadType* RMC_budgetFreeList;
extern  void* RMC_budgetFreeListEnd;

/*
 * UTILITY FUNCTIONS
 */

#define ASSERT_R( b ) \
  if ( !(b) ) { \
    printf( "ASSERT @ %s # %d, ", __FILE__, __LINE__); \
    printf("%s\n", resultToString(result)); \
  };

#define ASSERT( b , ...) \
  if ( !(b) ) { \
    printf( "ASSERT @ %s # %d, ", __FILE__, __LINE__); \
    printf(__VA_ARGS__); \
    printf("\n"); \
  };

char* resultToString(CR_Result result) {
  switch (result) {
    case CR_RESULT_OK :
      return "CR_RESULT_OK";
    case CR_ERROR_INVALID_IREF :
      return "CR_ERROR_INVALID_IREF";
    case CR_ERROR_INVALID_PARAM :
      return "CR_ERROR_INVALID_PARAM";
    case CR_ERROR_INVALID_STATE :
      return "CR_ERROR_INVALID_STATE";
      
    case CR_RMC_ERROR_RESOURCE :
      return "CR_RMC_ERROR_RESOURCE";
    case CR_RMC_ERROR_BUDGET :
      return "CR_RMC_ERROR_BUDGET";    
    case CR_RMC_ERROR_RMC :
      return "CR_RMC_ERROR_RMC";  
    case CR_RMC_ERROR_INVALID_HEAD_SIZE:
      return "CR_RMC_ERROR_INVALID_HEAD_SIZE";
    
    default :
      return "unknown result";
  }
  
  return "";
}

/* create a handle to a dummy object */
static CR_ObjectType* objectHandle = (void*)1;

/* RMC init parameters */
static CR_CreationParameterType rmcInitParams1[] =
{
  { RMC_PARAM_MEMORY_SIZE, "1024" },
  { RMC_PARAM_MAX_NUMBER_OF_BUDGETS, "1" },
  { CR_NULL, ""},
};

static CR_CreationParameterType rmcInitParams2[] =
{
  { RMC_PARAM_MEMORY_SIZE, "1024" },
  { RMC_PARAM_MAX_NUMBER_OF_BUDGETS, "2" },
  { CR_NULL, ""},
};

static CR_CreationParameterType rmcInitParams3[] =
{
  { RMC_PARAM_MEMORY_SIZE, "1024" },
  { RMC_PARAM_MAX_NUMBER_OF_BUDGETS, "3" },
  { CR_NULL, ""},
};

static CR_CreationParameterType rmcInitParams4[] =
{
  { RMC_PARAM_MEMORY_SIZE, "1024" },
  { RMC_PARAM_MAX_NUMBER_OF_BUDGETS, "4" },
  { CR_NULL, ""},
};

#define kTotalMemorySize 1024

RMC_BudgetHandleType createBudget(CR_UInt32 size, CR_Int32 allocations)
{
  RMC_BudgetHandleType budgetHandle;
  CR_Result result;

  struct RMC_MemoryBudgetRequestType budget;
  budget.allocationSize = size;
  budget.maxNumberOfAllocations = allocations;
  
  result = RMC_RequestMemoryBudget(objectHandle, &budget, &budgetHandle);
  
  /* sanity check */
  ASSERT_R(result == CR_RESULT_OK);
  
  return budgetHandle;
}

/*
 * TEST CASES
 */
 
/*
 * RMC_init()
 */
 
void test_RMC_init_OK() 
{
  CR_Result result; 
  result = RMC_init(rmcInitParams1);
  ASSERT_R(result == CR_RESULT_OK);
  
  result = RMC_finalize();
  ASSERT_R(result == CR_RESULT_OK);
}

/*
 * RMC_RequestMemoryBudget()
 */
 
void test_RMC_RequestMemoryBudget_OK() 
{
  RMC_MemoryBudgetHeadType* pBudget;
  CR_UInt32 N;
  
  RMC_init(rmcInitParams1);
    
  pBudget = (RMC_MemoryBudgetHeadType*)createBudget(100, 0);
  
  N = (kTotalMemorySize - RMC_BUDGET_HEAD_SIZE - 100);
  ASSERT (RMC_budgetFreeList->size == N, "%lu (expecting %lu)", RMC_budgetFreeList->size, N);

  RMC_finalize();
}

void test_RMC_RequestMemoryBudget_PARAMS() 
{
  RMC_BudgetHandleType budgetHandle;
  CR_Result result;
  
  /* create a handle to a dummy object */   
  CR_ObjectType handle = malloc(sizeof(CR_ObjectType));

  struct RMC_MemoryBudgetRequestType budget;    
  budget.allocationSize = 100;
  budget.maxNumberOfAllocations = 0;

  RMC_init(rmcInitParams3);

  result = RMC_RequestMemoryBudget(NULL, &budget, &budgetHandle);
  ASSERT_R(result == CR_ERROR_INVALID_PARAM);

  result = RMC_RequestMemoryBudget(handle, NULL, &budgetHandle);
  ASSERT_R(result == CR_ERROR_INVALID_PARAM);

  result = RMC_RequestMemoryBudget(handle, &budget, NULL);
  ASSERT_R(result == CR_ERROR_INVALID_PARAM);
  
  /* cleanup */
  free(handle);

  RMC_finalize();
}

void test_RMC_RequestMemoryBudget_TooMuch() 
{
  RMC_BudgetHandleType budgetHandle;
  CR_Result result;

  RMC_MemoryBudgetRequestType budget;    

  RMC_init(rmcInitParams1);
  createBudget(100, 0);

  budget.allocationSize = 1000;
  result = RMC_RequestMemoryBudget(objectHandle, &budget, &budgetHandle);

  ASSERT_R(result == CR_RMC_ERROR_BUDGET);
  
  RMC_finalize();
}

void test_RMC_RequestMemoryBudget_AlmostAll() 
{

  RMC_init(rmcInitParams2);
  createBudget(100, 0);

  createBudget(kTotalMemorySize - 100 - 3 * RMC_BUDGET_HEAD_SIZE, 0);
  
  RMC_finalize();
}

void test_RMC_RequestMemoryBudget_All() 
{
  RMC_init(rmcInitParams2);
  createBudget(100, 0);
  createBudget(kTotalMemorySize - 100 - 2 * RMC_BUDGET_HEAD_SIZE, 0);
  
  RMC_finalize();
}


/*
 * RMC_AllocateMemory()
 */

void test_RMC_AllocateMemory_OK() 
{
    RMC_BudgetHandleType budgetHandle;
    RMC_MemoryBudgetHeadType* pBudget;
    struct RMC_MemoryRequestType memory;    
    void* p = NULL;
    CR_Result result;

    RMC_init(rmcInitParams1);
    
    budgetHandle = createBudget(64, 1);
    memory.size = 16;
    memory.budget = budgetHandle;

    result = RMC_AllocateMemory(&memory, &p);
    ASSERT_R(result == CR_RESULT_OK);
    
    /* check whether the pointer is not null */
    ASSERT(p != NULL, "%p", p);
    
    /* check the remianing size of available budget */
    pBudget = (RMC_MemoryBudgetHeadType*)budgetHandle;
    ASSERT(pBudget->freeList->size == 64 - 16 - RMC_MEMORY_HEAD_SIZE, "%lu (expecting %u)", pBudget->freeList->size, 64 - 16 - RMC_MEMORY_HEAD_SIZE);
    ASSERT(pBudget->size == 64 + RMC_MEMORY_HEAD_SIZE, "%lu (expecting %lu)", pBudget->size, 64 + RMC_MEMORY_HEAD_SIZE);

    RMC_finalize();
}

void test_RMC_AllocateMemory_PARAMS() 
{
    RMC_BudgetHandleType budgetHandle;
    struct RMC_MemoryRequestType memory;
    void* p;
    CR_Result result;

    RMC_init(rmcInitParams1);

    budgetHandle = createBudget(100, 3);    
    memory.size = 100;
    memory.budget = budgetHandle;

    result = RMC_AllocateMemory(NULL, &p);
    ASSERT_R(result == CR_ERROR_INVALID_PARAM);

    result = RMC_AllocateMemory(&memory, NULL);
    ASSERT_R(result == CR_ERROR_INVALID_PARAM);

    memory.budget = NULL;
    result = RMC_AllocateMemory(&memory, &p);
    ASSERT_R(result == CR_ERROR_INVALID_PARAM);

    RMC_finalize();
}

void test_RMC_AllocateMemory_TooMuch()
{
    RMC_BudgetHandleType budgetHandle;
    RMC_MemoryBudgetHeadType* pBudget;
    void* p;
    CR_Result result;
    struct RMC_MemoryRequestType memory;    
    
    RMC_init(rmcInitParams1);

    budgetHandle = createBudget(RMC_UNIT_SIZE*10, 1);
    memory.size = RMC_UNIT_SIZE*10 + 1;
    memory.budget = budgetHandle;
    result = RMC_AllocateMemory(&memory, &p);
    ASSERT_R(result == CR_RMC_ERROR_BUDGET);

    /* the freeList should not be null, since the memory could not be allocated */
    pBudget = (RMC_MemoryBudgetHeadType*)budgetHandle;
    ASSERT(pBudget->freeList != NULL, "pBudget==NULL");
    
    RMC_finalize();
}

void test_RMC_AllocateMemory_All()
{
    RMC_BudgetHandleType budgetHandle;
    RMC_MemoryBudgetHeadType* pBudget;
    void* p;
    CR_Result result;
    struct RMC_MemoryRequestType memory;    
    CR_UInt32 availableBudget;
    
    RMC_init(rmcInitParams1);

    budgetHandle = createBudget(100, 1);
    memory.size = 100;
    memory.budget = budgetHandle;
    result = RMC_AllocateMemory(&memory, &p);
    ASSERT_R(result == CR_RESULT_OK);

    memory.size = 100;
    result = RMC_AllocateMemory(&memory, &p);
    ASSERT_R(result == CR_RMC_ERROR_BUDGET);
    
    RMC_getAvailableBudget(budgetHandle, &availableBudget);
    
    pBudget = (RMC_MemoryBudgetHeadType*)budgetHandle;
    ASSERT(pBudget->freeList == NULL, "%p (%lu)", (void *)pBudget->freeList, availableBudget);
    
    RMC_finalize();
}

void test_RMC_AllocateMemory_0() 
{
    RMC_BudgetHandleType budgetHandle;
    void* p;
    CR_Result result;
    struct RMC_MemoryRequestType memory;    
    
    RMC_init(rmcInitParams1);

    budgetHandle = createBudget(100, 1);
    memory.size = 0;
    memory.budget = budgetHandle;
    result = RMC_AllocateMemory(&memory, &p);
    ASSERT_R(result == CR_ERROR_INVALID_PARAM);

    RMC_finalize();
}

void test_RMC_AllocateMemory_multipleTooMuch() 
{
    RMC_BudgetHandleType budgetHandle;
    void* p;
    CR_Result result;
    struct RMC_MemoryRequestType memory;    
    
    RMC_init(rmcInitParams1);

    budgetHandle = createBudget(100, 4);
    memory.size = 50;
    memory.budget = budgetHandle;
    result = RMC_AllocateMemory(&memory, &p);
    ASSERT_R(result == CR_RESULT_OK);

    memory.size = RMC_MEMORY_HEAD_SIZE;
    result = RMC_AllocateMemory(&memory, &p);
    ASSERT_R(result == CR_RESULT_OK);

    memory.size = RMC_MEMORY_HEAD_SIZE;
    result = RMC_AllocateMemory(&memory, &p);
    ASSERT_R(result == CR_RESULT_OK);

    memory.size = 70;
    result = RMC_AllocateMemory(&memory, &p);
    ASSERT_R(result == CR_RMC_ERROR_BUDGET);

    RMC_finalize();
}

void test_RMC_AllocateMemory_multipleAll() 
{
    RMC_BudgetHandleType budgetHandle;
    void* p;
    CR_Result result;
    struct RMC_MemoryRequestType memory;    
    
    RMC_init(rmcInitParams1);
    
    budgetHandle = createBudget(100, 4);
    memory.size = 16;
    memory.budget = budgetHandle;
    result = RMC_AllocateMemory(&memory, &p);
    ASSERT_R(result == CR_RESULT_OK);

    memory.size = RMC_MEMORY_HEAD_SIZE;
    result = RMC_AllocateMemory(&memory, &p);
    ASSERT_R(result == CR_RESULT_OK);

    memory.size = 12;
    result = RMC_AllocateMemory(&memory, &p);
    ASSERT_R(result == CR_RESULT_OK);

    memory.size = 100 - 16 - RMC_MEMORY_HEAD_SIZE - 12 - 4*RMC_MEMORY_HEAD_SIZE;
    result = RMC_AllocateMemory(&memory, &p);
    ASSERT_R(result == CR_RESULT_OK);

    RMC_finalize();
}

/*
 * RMC_FreeMemory()
 */
 
void test_RMC_FreeMemory_OK()
{
  RMC_BudgetHandleType budgetHandle;
  void* p;
  CR_Result result;
  
  RMC_MemoryRequestType memory;    
  
  RMC_init(rmcInitParams1);

  budgetHandle = createBudget(100, 1);
  memory.budget = budgetHandle;
  memory.size = 10;
  result = RMC_AllocateMemory(&memory, &p);
  ASSERT_R(result == CR_RESULT_OK);

  result = RMC_FreeMemory(budgetHandle, p);
  ASSERT_R(result == CR_RESULT_OK);  
  

  RMC_finalize();
}

void test_RMC_FreeMemory_PARAM()
{
  RMC_BudgetHandleType budgetHandle;
  RMC_MemoryRequestType memory;    
  void *p;
  CR_Result result;
  
  RMC_init(rmcInitParams1);

  budgetHandle = createBudget(100, 1);
  memory.budget = budgetHandle;
  memory.size = 10;
  RMC_AllocateMemory(&memory, &p);
  
  result = RMC_FreeMemory(NULL, p);
  ASSERT_R(result == CR_ERROR_INVALID_PARAM);
  
  result = RMC_FreeMemory((void*)4, p);
  ASSERT_R(result == CR_ERROR_INVALID_PARAM);

  result = RMC_FreeMemory(budgetHandle, NULL);
  ASSERT_R(result == CR_ERROR_INVALID_PARAM);

  result = RMC_FreeMemory(budgetHandle, (void*)4);
  ASSERT_R(result == CR_ERROR_INVALID_PARAM);

  result = RMC_FreeMemory(budgetHandle, p);
  ASSERT_R(result == CR_RESULT_OK);

  RMC_finalize();
}

void test_RMC_FreeMemory_Multiple()
{
  RMC_BudgetHandleType budgetHandle;
  void *p, *q;
  CR_Result result;
  
  struct RMC_MemoryRequestType memory;    
  
  RMC_init(rmcInitParams1);
  
  budgetHandle = createBudget(100, 4);
  memory.budget = budgetHandle;
  memory.size = RMC_MEMORY_HEAD_SIZE;
  result = RMC_AllocateMemory(&memory, &p);
  ASSERT_R(result == CR_RESULT_OK);

  memory.size = 12;
  result = RMC_AllocateMemory(&memory, &p);
  ASSERT_R(result == CR_RESULT_OK);

  memory.size = 16;
  result = RMC_AllocateMemory(&memory, &q);
  ASSERT_R(result == CR_RESULT_OK);

  memory.size = 20;
  result = RMC_AllocateMemory(&memory, &p);
  ASSERT_R(result == CR_RESULT_OK);

  result = RMC_FreeMemory(budgetHandle, q);
  ASSERT_R(result == CR_RESULT_OK);  
  
  RMC_finalize();
}

void test_RMC_FreeMemory_Merge()
{
  RMC_BudgetHandleType budgetHandle;
  void *p, *q;
  CR_Result result;
  
  struct RMC_MemoryRequestType memory;    
  
  RMC_init(rmcInitParams1);

  budgetHandle = createBudget(100, 2);
  memory.budget = budgetHandle;
  
  memory.size = 12;
  result = RMC_AllocateMemory(&memory, &p);
  ASSERT_R(result == CR_RESULT_OK);  

  memory.size = 60;
  result = RMC_AllocateMemory(&memory, &q);
  ASSERT_R(result == CR_RESULT_OK);  

  result = RMC_FreeMemory(budgetHandle, p);
  ASSERT_R(result == CR_RESULT_OK);  

  result = RMC_FreeMemory(budgetHandle, q);
  ASSERT_R(result == CR_RESULT_OK);  
  
  memory.size = 92;
  result = RMC_AllocateMemory(&memory, &p);
  ASSERT_R(result == CR_RESULT_OK); 
  
  RMC_finalize();
}

void test_RMC_FreeMemory_Rewrite()
{
  RMC_BudgetHandleType budgetHandle;
  int *p, *q;
  CR_Result result;
  
  struct RMC_MemoryRequestType memory;
  
  RMC_init(rmcInitParams1);

  budgetHandle = createBudget(100, 1);
  memory.budget = budgetHandle;
  memory.size = 92;
  result = RMC_AllocateMemory(&memory, (void*)&p);
  ASSERT_R(result == CR_RESULT_OK);  

  *p = 123;

  result = RMC_FreeMemory(budgetHandle, p);
  ASSERT_R(result == CR_RESULT_OK);
  
  memory.size = 92;
  result = RMC_AllocateMemory(&memory, (void*)&q);
  ASSERT_R(result == CR_RESULT_OK);
  
  *q = 456;
  ASSERT_R(*p == *q);

  RMC_finalize();
}

/*
 * RMC_DiscardBudget()
 */
 
void test_RMC_DiscardBudget_OK()
{
  RMC_BudgetHandleType budgetHandle;
  CR_Result result;
  
  RMC_init(rmcInitParams1);

  budgetHandle = createBudget(100, 0);
  result = RMC_DiscardMemoryBudget(budgetHandle);
  ASSERT_R(result == CR_RESULT_OK);

  RMC_finalize();
}

void test_RMC_DiscardBudget_PARAMS()
{
  CR_Result result;

  RMC_init(rmcInitParams1);

  result = RMC_DiscardMemoryBudget(NULL);
  ASSERT_R(result == CR_ERROR_INVALID_PARAM);

  result = RMC_DiscardMemoryBudget((void*)100);
  ASSERT_R(result == CR_ERROR_INVALID_PARAM);
  
  RMC_finalize();
}

void test_RMC_DiscardBudget_Multiple()
{
  RMC_BudgetHandleType a, b, c, d;
  CR_Result result;
  
  RMC_init(rmcInitParams4);

  a = createBudget(100, 0);
  b = createBudget(100, 0);
  c = createBudget(100, 0);
  d = createBudget(100, 0);

  result = RMC_DiscardMemoryBudget(b);
  ASSERT_R(result == CR_RESULT_OK);

  result = RMC_DiscardMemoryBudget(d);
  ASSERT_R(result == CR_RESULT_OK);

  result = RMC_DiscardMemoryBudget(a);
  ASSERT_R(result == CR_RESULT_OK);

  result = RMC_DiscardMemoryBudget(c);
  ASSERT_R(result == CR_RESULT_OK);

  RMC_finalize();
}

void test_RMC_DiscardBudget_All()
{
  RMC_BudgetHandleType a, b, c, d;
  CR_Result result;
  
  RMC_init(rmcInitParams4);

  a = createBudget(100, 0);
  b = createBudget(100, 0);
  c = createBudget(100, 0);
  d = createBudget(1024 - 4*RMC_BUDGET_HEAD_SIZE - 3*100, 0);

  result = RMC_DiscardMemoryBudget(b);
  ASSERT_R(result == CR_RESULT_OK);

  result = RMC_DiscardMemoryBudget(d);
  ASSERT_R(result == CR_RESULT_OK);

  result = RMC_DiscardMemoryBudget(a);
  ASSERT_R(result == CR_RESULT_OK);

  result = RMC_DiscardMemoryBudget(c);
  ASSERT_R(result == CR_RESULT_OK);

  RMC_finalize();
}

void test_RMC_DiscardBudget_Merge()
{
  RMC_BudgetHandleType a, b, c, d;
  CR_Result result;
  
  RMC_init(rmcInitParams4);

  a = createBudget(100, 0);
  b = createBudget(100, 0);
  c = createBudget(100, 0);
  d = createBudget(1024 - 4*RMC_BUDGET_HEAD_SIZE - 3*100, 0);

  result = RMC_DiscardMemoryBudget(b);
  ASSERT_R(result == CR_RESULT_OK);

  result = RMC_DiscardMemoryBudget(d);
  ASSERT_R(result == CR_RESULT_OK);

  result = RMC_DiscardMemoryBudget(a);
  ASSERT_R(result == CR_RESULT_OK);

  result = RMC_DiscardMemoryBudget(c);
  ASSERT_R(result == CR_RESULT_OK);

  createBudget(1024-RMC_BUDGET_HEAD_SIZE, 0);

  RMC_finalize();
}

void test_RMC_DiscardBudget_Rewrite()
{
  RMC_BudgetHandleType a, b;
  RMC_MemoryRequestType memory;
  CR_Result result;
  int *p, *q;

  RMC_init(rmcInitParams4);

  createBudget(100, 1);
  a = createBudget(100, 1);  
  createBudget(100, 1);
  createBudget(1024 - 4*RMC_BUDGET_HEAD_SIZE - 3*100, 1);
  
  memory.budget = a;
  memory.size = sizeof(int);
  result = RMC_AllocateMemory(&memory, (void*)&p);
  ASSERT_R(result == CR_RESULT_OK);
  
  *p = 123;
  
  result = RMC_DiscardMemoryBudget(a);
  ASSERT_R(result == CR_RESULT_OK);

  b = createBudget(100, 1);  

  result = RMC_AllocateMemory(&memory, (void*)&q);
  ASSERT_R(result == CR_RESULT_OK);
  
  *q = 123;
  
  ASSERT_R(*p == *q);
  
  RMC_finalize();
}


/*
 * RMC_ResetMemoryBudget()
 */

void test_RMC_ResetMemoryBudget_OK()
{
  RMC_BudgetHandleType budgetHandle;
  void* p1;
  void* p2;
  char* string1;
  char* string2;
  CR_Result result;
  struct RMC_MemoryRequestType memory;    
  
  RMC_init(rmcInitParams1);

  budgetHandle = createBudget(100, 1);
  memory.size = 70;
  memory.budget = budgetHandle;
  result = RMC_AllocateMemory(&memory, &p1);
  ASSERT_R(result == CR_RESULT_OK);

  /* write to the allocated memory */
  string1 = "TestString";
  strcpy(p1, string1);
  
  /* reset the budget */
  result = RMC_ResetMemoryBudget(budgetHandle);
  ASSERT_R(result == CR_RESULT_OK);
  
  /* allocate memory again, which would spill over if budget was not reset.
     make sure same allocation size is used, because allocated form teh end */
  memory.size = 70;
  
  result = RMC_AllocateMemory(&memory, &p2);
  ASSERT_R(result == CR_RESULT_OK);
  
  /* write to the allocated memory */
  string2 = "StringTest";
  strcpy(p2, string2);
  
  /* check if the new string was written to the old value of p */
  ASSERT(strcmp(p1, p2) == 0, "%s != %s (%p != %p)", (char*)p1, (char*)p2, p1, p2);

  RMC_finalize();
}

void test_RMC_ResetMemoryBudget_PARAM()
{
  CR_Result result;
  
  RMC_init(rmcInitParams1);

  result = RMC_ResetMemoryBudget(NULL);
  ASSERT_R(result == CR_ERROR_INVALID_PARAM);

  result = RMC_ResetMemoryBudget((void*)100);
  ASSERT_R(result == CR_ERROR_INVALID_PARAM);

  RMC_finalize();
}

/*
 * RMC_getAvailableBudget()
 */

void test_RMC_getAvailableBudget_OK()
{
  RMC_BudgetHandleType budgetHandle;
  struct RMC_MemoryRequestType memory;    
  void* p1;
  CR_Result result;
  CR_UInt32 expected, utilization;
    
  RMC_init(rmcInitParams1);

  budgetHandle = createBudget(100, 1);
  memory.budget = budgetHandle;

  memory.size = 72;
  result = RMC_AllocateMemory(&memory, &p1);
  ASSERT_R(result == CR_RESULT_OK);

  expected = 100 - memory.size;
  result = RMC_getAvailableBudget(budgetHandle, &utilization);
  ASSERT_R(result == CR_RESULT_OK);
  ASSERT(utilization == expected, "unexpected utilization %lu, expecting %lu", utilization, expected);
  
  RMC_finalize();
}

void test_RMC_getAvailableBudget_PARAMS()
{
  RMC_BudgetHandleType budgetHandle;
  RMC_MemoryRequestType memory;    
  void *p1;
  CR_Result result;
  CR_UInt32 utilization;
    
  RMC_init(rmcInitParams1);

  budgetHandle = createBudget(100, 1);
  memory.size = 100;
  memory.budget = budgetHandle;
  result = RMC_AllocateMemory(&memory, &p1);
  ASSERT_R(result == CR_RESULT_OK);

  result = RMC_getAvailableBudget(NULL, &utilization);
  ASSERT_R(result == CR_ERROR_INVALID_PARAM);

  result = RMC_getAvailableBudget(budgetHandle, NULL);
  ASSERT_R(result == CR_ERROR_INVALID_PARAM);

  result = RMC_getAvailableBudget((RMC_BudgetHandleType*)p1, NULL);
  ASSERT_R(result == CR_ERROR_INVALID_PARAM);
  
  RMC_finalize();
}

void test_RMC_getAvailableBudget_offset()
{
  RMC_BudgetHandleType budgetHandle;
  struct RMC_MemoryRequestType memory;    
  void* p1;
  CR_Result result;
  CR_UInt32 expected, utilization;
    
  RMC_init(rmcInitParams1);

  budgetHandle = createBudget(100, 1);
  memory.budget = budgetHandle;

  memory.size = 70;
  result = RMC_AllocateMemory(&memory, &p1);
  ASSERT_R(result == CR_RESULT_OK);

  expected = 100 - memory.size - 2;
  result = RMC_getAvailableBudget(budgetHandle, &utilization);
  ASSERT_R(result == CR_RESULT_OK);
  ASSERT(utilization == expected, "unexpected utilization %lu, expecting %lu", utilization, expected);
  
  RMC_finalize();
}

void test_RMC_getAvailableBudget_multiple()
{
  RMC_BudgetHandleType budgetHandle;
  struct RMC_MemoryRequestType memory;    
  void* p1;
  CR_Result result;
  CR_UInt32 expected, utilization;
    
  RMC_init(rmcInitParams1);

  budgetHandle = createBudget(100, 2);
  memory.budget = budgetHandle;

  memory.size = 40;
  result = RMC_AllocateMemory(&memory, &p1);
  ASSERT_R(result == CR_RESULT_OK);

  expected = 100 - (memory.size);
  result = RMC_getAvailableBudget(budgetHandle, &utilization);
  ASSERT_R(result == CR_RESULT_OK);
  ASSERT(utilization == expected, "unexpected utilization %lu, expecting %lu", utilization, expected);

  memory.size = RMC_MEMORY_HEAD_SIZE;
  result = RMC_AllocateMemory(&memory, &p1);
  ASSERT_R(result == CR_RESULT_OK);

  expected = expected - (memory.size);
  result = RMC_getAvailableBudget(budgetHandle, &utilization);
  ASSERT_R(result == CR_RESULT_OK);
  ASSERT(utilization == expected, "unexpected utilization %lu, expecting %lu", utilization, expected);
  
  RMC_finalize();
}

void test_RMC_getAvailableBudget_extend()
{
  RMC_BudgetHandleType budgetHandle;
  struct RMC_MemoryRequestType memory;    
  void* p1;
  CR_Result result;
  CR_UInt32 expected, utilization;
    
  RMC_init(rmcInitParams1);

  budgetHandle = createBudget(100, 2);
  memory.budget = budgetHandle;

  memory.size = 72;
  result = RMC_AllocateMemory(&memory, &p1);
  ASSERT_R(result == CR_RESULT_OK);

  expected = 100 - memory.size;
  result = RMC_getAvailableBudget(budgetHandle, &utilization);
  ASSERT_R(result == CR_RESULT_OK);
  ASSERT(utilization == expected, "unexpected utilization %lu, expecting %lu", utilization, expected);

  memory.size = 100 - memory.size;
  result = RMC_AllocateMemory(&memory, &p1);
  ASSERT_R(result == CR_RESULT_OK);

  expected = 0;
  result = RMC_getAvailableBudget(budgetHandle, &utilization);
  ASSERT_R(result == CR_RESULT_OK);
  ASSERT(utilization == expected, "unexpected utilization %lu, expecting %lu", utilization, expected);
  
  RMC_finalize();
}

void test_RMC_getAvailableBudget_free()
{
  RMC_BudgetHandleType budgetHandle;
  RMC_MemoryRequestType memory1, memory2;    
  void *p1, *p2;
  CR_Result result;
  CR_UInt32 expected, utilization;
    
  RMC_init(rmcInitParams1);

  budgetHandle = createBudget(100, 2);
  memory1.budget = budgetHandle;

  memory1.size = 40;
  result = RMC_AllocateMemory(&memory1, &p1);
  ASSERT_R(result == CR_RESULT_OK);

  expected = 100 - memory1.size;
  result = RMC_getAvailableBudget(budgetHandle, &utilization);
  ASSERT_R(result == CR_RESULT_OK);
  ASSERT(utilization == expected, "unexpected utilization %lu, expecting %lu", utilization, expected);

  memory2.budget = budgetHandle;
  memory2.size = RMC_MEMORY_HEAD_SIZE;
  result = RMC_AllocateMemory(&memory2, &p2);
  ASSERT_R(result == CR_RESULT_OK);

  expected = expected - (memory2.size);
  result = RMC_getAvailableBudget(budgetHandle, &utilization);
  ASSERT_R(result == CR_RESULT_OK);
  ASSERT(utilization == expected, "unexpected utilization %lu, expecting %lu", utilization, expected);
  
  result = RMC_FreeMemory(budgetHandle, p1);
  ASSERT_R(result == CR_RESULT_OK);

  expected = expected + (memory1.size);
  result = RMC_getAvailableBudget(budgetHandle, &utilization);
  ASSERT_R(result == CR_RESULT_OK);
  ASSERT(utilization == expected, "unexpected utilization %lu, expecting %lu", utilization, expected);
  
  RMC_finalize();
}

void test_RMC_RoundBytesToWords_Equal() {
  CR_UInt32 given, rounded, expected;
  
  given = 3*sizeof(void*);
  expected = 3*sizeof(void*);
  rounded = RMC_RoundBytesToWords(given);
  ASSERT(rounded == expected, "unexpected number of bytes %lu, expecting %lu", rounded, expected);
}

void test_RMC_RoundBytesToWords_Different() {
  CR_UInt32 given, rounded, expected;
  
  given = 3*sizeof(void*)+1;
  expected = 4*sizeof(void*);
  rounded = RMC_RoundBytesToWords(given);
  ASSERT(rounded == expected, "unexpected number of bytes %lu, expecting %lu", rounded, expected);
}


/* 
 * run tests
 */

void test_RMC_init(void) {
  test_RMC_init_OK();
  printf("FINISHED test_RMC_init\n");
}

void test_RMC_RequestMemoryBudget(void) {
  test_RMC_RequestMemoryBudget_OK();
  test_RMC_RequestMemoryBudget_PARAMS();
  test_RMC_RequestMemoryBudget_TooMuch();
  test_RMC_RequestMemoryBudget_AlmostAll();
  test_RMC_RequestMemoryBudget_All();
  printf("FINISHED test_RMC_RequestMemoryBudge\n");
}

void test_RMC_AllocateMemory(void) {
  test_RMC_AllocateMemory_OK();
  test_RMC_AllocateMemory_PARAMS();
  test_RMC_AllocateMemory_TooMuch();
  test_RMC_AllocateMemory_All();
  test_RMC_AllocateMemory_0();
  test_RMC_AllocateMemory_multipleTooMuch();
  test_RMC_AllocateMemory_multipleAll();
  printf("FINISHED test_RMC_AllocateMemory\n");
}

void test_RMC_FreeMemory(void) {
  test_RMC_FreeMemory_OK();
  test_RMC_FreeMemory_PARAM();
  test_RMC_FreeMemory_Multiple();
  test_RMC_FreeMemory_Merge();
  test_RMC_FreeMemory_Rewrite();
  printf("FINISHED test_RMC_FreeMemory\n");
}

void test_RMC_DiscardBudge(void) {
  test_RMC_DiscardBudget_OK();
  test_RMC_DiscardBudget_PARAMS();
  test_RMC_DiscardBudget_Multiple();
  test_RMC_DiscardBudget_All();
  test_RMC_DiscardBudget_Merge();
  test_RMC_DiscardBudget_Rewrite();
  printf("FINISHED test_RMC_DiscardBudget\n");
}

void test_RMC_ResetMemoryBudge(void) {
  test_RMC_ResetMemoryBudget_OK();
  test_RMC_ResetMemoryBudget_PARAM();
  printf("FINISHED test_RMC_ResetMemoryBudget\n");
}

void test_RMC_getAvailableBudget(void) {
  test_RMC_getAvailableBudget_OK();
  test_RMC_getAvailableBudget_PARAMS();
  test_RMC_getAvailableBudget_offset();
  test_RMC_getAvailableBudget_multiple();
  test_RMC_getAvailableBudget_extend();
  test_RMC_getAvailableBudget_free();
  printf("FINISHED test_RMC_getAvailableBudget\n");
}

void test_RMC_RoundBytesToWords(void) {
  test_RMC_RoundBytesToWords_Equal();
  test_RMC_RoundBytesToWords_Different();
  printf("FINISHED test_RMC_RoundBytesToWords\n");
}

void test_RMC (void) {
  test_RMC_init();
  test_RMC_RequestMemoryBudget();
  test_RMC_AllocateMemory();
  test_RMC_FreeMemory();
  test_RMC_DiscardBudge();
  test_RMC_ResetMemoryBudge();
  test_RMC_getAvailableBudget();
  test_RMC_RoundBytesToWords();
  printf("FINISHED test_RMC\n");
}
